from collections import defaultdict

from elasticsearch_dsl import Search
from rest_framework.filters import BaseFilterBackend


class BaseESFilterBackend(BaseFilterBackend):
    def filter_queryset(self, request, queryset, view):
        return queryset

    def process_result(self, request, pagination, paginator, view):
        pass


class ESAggsFilterBackend(BaseESFilterBackend):
    def filter_queryset(self, request, queryset, view):
        for agg in view.aggs:
            agg = agg.top
            agg.apply(queryset.aggs)
        if hasattr(view, "parse_facet_query"):
            f = view.parse_facet_query(request)
        else:
            f = request.GET.get("f", None)
            if f:
                f = self.parse_facets(f)
        if f:
            for agg in view.aggs:
                agg = agg.top
                for query in agg.filter(f):
                    queryset = queryset.filter(query)
        return queryset

    def parse_facets(self, facets):
        facets = [
            x.replace("%3A", ":").replace("%3a", ":").replace("%25", "%")
            for x in facets.split(":")
        ]
        ret = defaultdict(list)
        for k, v in zip(facets[0::2], facets[1::2]):
            ret[k].append(v)
        return ret

    def process_result(self, request, pagination, paginator, view):
        agg_result = paginator.aggs.to_dict()
        ret_aggs = []
        for agg in view.aggs:
            agg = agg.top
            r = agg.process_result(agg_result)
            if r:
                ret_aggs.append(r)
        paginator.aggs = ret_aggs


class QueryFilterBackend(BaseESFilterBackend):
    def filter_queryset(self, request, queryset, view):
        q = request.GET.get("q")
        if not q:
            return queryset
        interpreter = self._get_query_interpreter(view, request)
        if interpreter:
            queryset = interpreter(request, queryset, view, q)
        else:
            return queryset.filter("match_none")  # match none on invalid parser
        return queryset

    @staticmethod
    def _get_query_interpreter(view, request):
        parser_id = request.GET.get("parser", None)
        if not parser_id:
            return view.default_query_interpreter
        return view.query_interpreters.get(parser_id)



from elasticsearch_dsl import Q


class AggBase:
    insert_into_parent = False

    def __init__(
        self, code, nested=None, parent=None, label=None, display=True, **kwargs
    ):
        self.code = code
        self.label = label
        self.parent = parent
        self.display = display
        self.nested = list(nested or ())
        self.kwargs = kwargs

    def apply(self, agg):
        if self.display:
            for n in self.nested:
                n.apply(agg)
        return agg

    def _append(self, item_type, *args, **kwargs):
        if self.insert_into_parent and self.parent:
            return self.parent._append(item_type, *args, **kwargs)
        b = item_type(*args, **kwargs)
        self.nested.append(b)
        b.parent = self
        return b

    def bucket(self, *args, **kwargs):
        return self._append(BucketAgg, *args, **kwargs)

    def separator(self):
        self._append(SeparatorAgg)
        return self

    @property
    def top(self):
        if self.parent:
            return self.parent.top
        return self

    def process_result(self, data):
        if self.code not in data:
            return None

        my_data = data.pop(self.code)

        ret = {"code": self.code}
        missing_key = self.code + ".__missing__"
        if missing_key in data and "doc_count" in data[missing_key]:
            ret["__missing__"] = data.pop(missing_key)["doc_count"]

        count_key = self.code + ".__count__"
        if count_key in data and "value" in data[count_key]:
            ret["__count__"] = data.pop(count_key)["value"]

        if self.label:
            ret["label"] = self.label

        if self.nested:
            sub_facets = []
            for n in self.nested:
                r = n.process_result(my_data)
                if r:
                    sub_facets.append(r)
            ret["facets"] = sub_facets
        ret.update(my_data)
        return ret

    def filter(self, facets_and_values):
        if self.nested:
            for n in self.nested:
                yield from n.filter(facets_and_values)


class BucketAgg(AggBase):
    insert_into_parent = True

    def __init__(self, code, type="terms", field=None, **kwargs):
        super().__init__(code, **kwargs)
        self.type = type
        self.field = field or code

    def apply(self, agg):
        if self.display:
            agg.bucket(
                self.code + ".__missing__", "missing", field=self.field, **self.kwargs
            )
            agg.bucket(
                self.code + ".__count__", "value_count", field=self.field, **self.kwargs
            )
            ret = agg.bucket(self.code, self.type, field=self.field, **self.kwargs)
            return super().apply(ret)
        else:
            return agg

    def filter(self, facets_and_values):
        values = self.get_values(facets_and_values)
        if values:
            not_missing = [
                x for x in values if x != "--missing--" and x != "--exists--"
            ]
            missing = [x for x in values if x == "--missing--"]
            exists = [x for x in values if x == "--exists--"]
            if not_missing:
                yield Q("terms", **{self.field: values})
            if missing:
                yield Q("bool", must_not=Q("exists", field=self.field))
            if exists:
                yield Q("exists", field=self.field)
        yield from super().filter(facets_and_values)

    def get_values(self, facets_and_values):
        return facets_and_values.get(self.code, None)


class TranslatedBucketAgg(BucketAgg):
    def __init__(self, code, type="terms", field=None, map=None, **kwargs):
        super().__init__(code, type=type, field=field, **kwargs)
        self.map = map or {}
        self.inverse_map = {v: k for k, v in self.map.items()}

    def get_values(self, facets_and_values):
        ret = super().get_values(facets_and_values)
        if ret:
            ret = [self.inverse_map.get(x, x) for x in ret]
        return ret

    def process_result(self, data):
        ret = super().process_result(data)
        if "buckets" in ret:
            ret["buckets"] = [
                {
                    **x,
                    "key": self.map.get(x["key"], None)
                    or self.map.get("key_as_string", None)
                    or x["key"],
                }
                for x in ret["buckets"]
            ]
        return ret


class NestedAgg(AggBase):
    def __init__(self, code, path=None, **kwargs):
        super().__init__(code, **kwargs)
        self.path = path or code

    def apply(self, agg):
        if self.display:
            ret = agg.bucket(self.code, "nested", path=self.path)
            return super().apply(ret)
        else:
            return agg

    def filter(self, facets_and_values):
        children = list(super().filter(facets_and_values))
        if children:
            yield Q("nested", path=self.path, query=Q("bool", must=children))



from functools import partial

from django.http import Http404
from django.shortcuts import get_object_or_404
from django.utils.functional import classproperty
from rest_framework import viewsets
from rest_framework.renderers import BrowsableAPIRenderer

from .backends.filters import ESAggsFilterBackend, QueryFilterBackend
from .backends.query_interpreters import (
    simple_query_interpreter,
    luqum_query_interpreter,
)
from .pagination import ESPagination
from .renderers import ESRenderer
from .serializers import ESDocumentSerializer
from .backends.source import DynamicSourceBackend
from .. import registry


class ESViewSet(viewsets.ModelViewSet):
    pagination_class = ESPagination
    renderer_classes = [
        ESRenderer,
        BrowsableAPIRenderer,
    ]
    aggs = ()
    filter_backends = [DynamicSourceBackend, ESAggsFilterBackend, QueryFilterBackend]
    serializer_class = ESDocumentSerializer
    query_interpreters = {
        "simple": simple_query_interpreter,
        "luqum": luqum_query_interpreter,
    }
    # needed so that the function is not bound (otherwise it would get 'self' as the first parameter)
    default_query_interpreter = staticmethod(simple_query_interpreter)
    lookup_url_kwarg = "pk"

    @property
    def document(self):
        raise AttributeError(
            f"Please specify document=<document class> on this viewset ({type(self)})"
        )

    @classproperty
    def lookup_field(cls):
        return cls.document.DOCUMENT_ID_FIELD

    def get_queryset(self):
        return self.document.search()

    def get_object(self):
        queryset = self.filter_queryset(self.get_queryset())
        lookup_url_kwarg = self.lookup_url_kwarg or self.lookup_field
        # TODO: remove aggs from the queryset
        _id = self.kwargs[lookup_url_kwarg]
        obj_list = queryset.filter("term", _id=_id)[:1].execute()
        count = obj_list.hits.total.value
        if not count:
            raise Http404(f"No document of type {self.document} with id {_id}")
        hit = obj_list.hits.hits[0].to_dict()
        obj = self.document.from_es(hit)
        # May raise a permission denied
        self.check_object_permissions(self.request, obj)

        return obj

    def get_serializer(self, *args, **kwargs):
        kwargs["django_serializer"] = registry.get_registry_entry_from_document(
            self.document
        ).serializer
        return super().get_serializer(*args, **kwargs)



from collections import namedtuple
from dataclasses import dataclass
from typing import Type, Dict

from elasticsearch_dsl import Document
from django.db.models import Model
from rest_framework.serializers import ModelSerializer

from . import settings
from .drf.serializers import StrictModelSerializer
from .indexer import disabled_es
from .json import to_plain_json

from django.db.models.signals import post_save, post_delete

import lazy_object_proxy


@dataclass
class DocumentRegistryEntry:
    model: Type[Model]
    document: Type[Document]
    serializer: Type[ModelSerializer]


class DjangoDocument(Document):
    DOCUMENT_ID_FIELD = "id"
    DJANGO_ID_FIELD = "pk"

    def save(self, **kwargs):
        with disabled_es():
            # do not save to elasticsearch inside this block as we will call super below
            obj = registry.save_to_django_object(self)
            obj.save()
        _id = getattr(obj, self.DJANGO_ID_FIELD)
        setattr(self, self.DOCUMENT_ID_FIELD, _id)
        self.meta._id = _id
        self.meta.id = _id
        if "refresh" not in kwargs:
            kwargs["refresh"] = True
        return super().save(**kwargs)

    def delete(self, using=None, index=None, **kwargs):
        entry = registry.get_registry_entry_from_document(type(self))
        obj = entry.model.objects.get(**{self.DJANGO_ID_FIELD: self.meta.id})
        with disabled_es():
            obj.delete()
        if "refresh" not in kwargs:
            kwargs["refresh"] = True
        return super().delete(using=using, index=index, **kwargs)

    @classmethod
    def from_django(cls, pk):
        return registry.load_from_django_object(cls, pk)

    def to_django(self):
        return registry.save_to_django_object(self)


RegistrationContext = namedtuple(
    "RegistrationContext", "model, serializer, mapping, prefix, included, excluded"
)


class DocumentRegistry:
    def __init__(self):
        self.by_model = {}
        self.by_document = {}

    def register(
        self,
        model: Type[Model],
        serializer: Type[ModelSerializer] = None,
        serializer_meta: Dict[str, any] = None,
        generate=True,
        included=(),
        excluded=(),
        mapping=None,
        serializer_mapping=None,
    ):
        """
        A decorator that registers a model with a DSL document. Usage

        @registry.register(MyModel)
        class MyModelDocument(DjangoDocument):
            class Index:
                name="blah"

        :param model:           django model
        :param serializer:      optional DRF serializer to turn the model into document's data
        :param serializer_meta: if serializer is not passed and is generated automatically, this
                                will be used to extend the meta of the serializer
        :param generate:        automatically generate fields not present on the document
        :param included:        list of fields to include in the document. Empty means all fields
        :param excluded:        list of fields to exclude from the document. Empty means all fields
        :param mapping:         a dictionary that specifies the mapping between DRF serializer fields
                                and Document fields. There are multiple cases
                                * key is a DRF field class, value a Document field class
                                * key is a DRF field class, value a callable(field: DRFField, context: RegistrationContext)
                                  that returns a Document field class or directly a Document field instance
                                * key is a string with field name(with optional dot notation), value a Document field
                                  class or a callable as above
                                * key is a string with field name (with optional dot notation), value a dict of parameters
                                  that will be added to field's constructor
        :return:                decorator
        """
        mapping = {**settings.DJANGO_ES_DEFAULT_FIELD_MAPPING, **(mapping or {})}
        serializer_mapping = {
            **settings.ES_DRF_DEFAULT_FIELD_MAPPING,
            **(serializer_mapping or {}),
        }

        if serializer is None:
            # generate serializer if not passed
            serializer = type(
                f"{model.__name__}Serializer",
                (StrictModelSerializer,),
                {
                    "Meta": type(
                        "Meta",
                        (),
                        {"model": model, "exclude": (), **(serializer_meta or {})},
                    )
                },
            )

        def do_registration(document, model, serializer):
            entry = DocumentRegistryEntry(
                model=model, document=document, serializer=serializer
            )
            self.by_model[model] = entry
            self.by_document[document] = entry

            from .signals import data_saved, data_deleted

            post_save.connect(data_saved, sender=model)
            post_delete.connect(data_deleted, sender=model)

        def delayed_registration(document, model, serializer):
            from .document_generator import generate_extra_document_fields
            from django_es_drf.serializer_generator import (
                generate_extra_serializer_fields,
            )

            document = generate_extra_document_fields(
                document, model, serializer, mapping, included, excluded
            )
            serializer = generate_extra_serializer_fields(
                document, model, serializer, serializer_mapping, included, excluded
            )
            do_registration(document, model, serializer)
            return document

        def wrapper(document: Type[Document]):
            if generate:
                return lazy_object_proxy.Proxy(
                    lambda: delayed_registration(document, model, serializer)
                )
            else:
                do_registration(document, model, serializer)
                return document

        return wrapper

    def save_to_django_object(self, document, django_object=None):
        entry = self.get_registry_entry_from_document(type(document))
        data = to_plain_json(document)
        try:
            id = getattr(document, document.DOCUMENT_ID_FIELD)
        except:
            id = None
        if django_object is None:
            if id is not None:
                django_object = entry.model.objects.filter(
                    **{document.DJANGO_ID_FIELD: id}
                ).first()
        serializer = entry.serializer(data=data)
        serializer.is_valid(True)

        if not django_object:
            return serializer.create(serializer.validated_data)
        else:
            return serializer.update(django_object, serializer.validated_data)

    def load_from_django_object(self, documentType: Type[Document], obj_or_pk):
        entry = self.get_registry_entry_from_document(documentType)
        if not isinstance(obj_or_pk, entry.model):
            obj = entry.model.objects.get(**{documentType.DJANGO_ID_FIELD: obj_or_pk})
        else:
            obj = obj_or_pk
        serializer = entry.serializer(instance=obj)
        return documentType(
            meta={
                "id": getattr(obj, documentType.DJANGO_ID_FIELD),
            },
            **serializer.data,
        )

    def model_to_index_and_id_and_data(self, obj):
        entry = self.get_registry_entry_from_django(type(obj))
        doc = self.load_from_django_object(entry.document, obj)
        return doc._index._name, doc.meta.id, to_plain_json(doc)

    def get_registry_entry_from_document(
        self, document: Type[Document]
    ) -> DocumentRegistryEntry:
        for m in document.mro():
            if m in self.by_document:
                return self.by_document[m]
        raise KeyError(
            f"Document of type {document} not found in django document registry"
        )

    def get_registry_entry_from_django(
        self, django: Type[Model]
    ) -> DocumentRegistryEntry:
        for m in django.mro():
            if m in self.by_model:
                return self.by_model[m]
        raise KeyError(
            f"Django object of type {django} not found in django document registry"
        )


registry = DocumentRegistry()



import math
from collections import OrderedDict

from django.core.paginator import Paginator, PageNotAnInteger, EmptyPage, Page
from django.utils.translation import gettext_lazy as _
from rest_framework.pagination import PageNumberPagination
from rest_framework.response import Response


class ESPage(Page):
    pass


class ESPaginator(Paginator):
    def validate_number(self, number):
        """Validate the given 1-based page number."""
        try:
            if isinstance(number, float) and not number.is_integer():
                raise ValueError
            number = int(number)
        except (TypeError, ValueError):
            raise PageNotAnInteger(_("That page number is not an integer"))
        if number < 1:
            raise EmptyPage(_("That page number is less than 1"))
        return number

    def page(self, number):
        """Return a Page object for the given 1-based page number."""
        number = self.validate_number(number)
        bottom = (number - 1) * self.per_page
        top = bottom + self.per_page
        self.object_list = self.object_list[bottom:top].execute()
        self.total = self.object_list.hits.total.value
        return ESPage(self.object_list, number, self)

    @property
    def num_pages(self):
        return math.ceil(self.total / self.per_page)

    @property
    def aggs(self):
        if hasattr(self, "_aggs") and self._aggs:
            return self._aggs
        return self.object_list.aggs

    @aggs.setter
    def aggs(self, value):
        self._aggs = value


class ESPagination(PageNumberPagination):
    django_paginator_class = ESPaginator
    page_size_query_param = "size"
    page_size = 10

    def get_paginated_response(self, data):
        for backend in self.view.filter_backends:
            if hasattr(backend, "process_result"):
                backend().process_result(
                    self.request, self, self.page.paginator, self.view
                )

        return Response(
            OrderedDict(
                [
                    ("count", self.page.paginator.total),
                    ("page", self.page.number),
                    ("size", self.page.paginator.per_page),
                    ("pages", self.page.paginator.num_pages),
                    ("next", self.get_next_link()),
                    ("previous", self.get_previous_link()),
                    ("hits", data),
                    ("aggs", self.page.paginator.aggs),
                ]
            )
        )

    def paginate_queryset(self, queryset, request, view=None):
        self.view = view
        return super().paginate_queryset(queryset, request, view)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['django_es_drf', 'django_es_drf.drf', 'django_es_drf.drf.backends']

package_data = \
{'': ['*']}

install_requires = \
['django>=4,<5',
 'djangorestframework>=3.13.1,<4.0.0',
 'elasticsearch-dsl>=7,<8',
 'lazy-object-proxy>=1.7.1,<2.0.0',
 'luqum>=0.11.0,<0.12.0']

setup_kwargs = {
    'name': 'django-es-drf',
    'version': '1.0.0a5',
    'description': 'Simple django - elasticsearch - drf integration',
    'long_description': '# Django ES DRF\n\n**Work in progress**\n\nA simple integration layer between Django, Elasticsearch and Django rest framework\n\n- [Django ES DRF](#django-es-drf)\n    - [Model and ES Document example](#model-and-es-document-example)\n    - [DRF example](#drf-example)\n        - [Returned fields from listing](#returned-fields-from-listing)\n        - [Search, facets](#search-facets)\n            - [Search via "q" GET parameters](#search-via-q-get-parameters)\n                - [Simple search](#simple-search)\n                - [Luqum search](#luqum-search)\n            - [Facets and faceted filtering](#facets-and-faceted-filtering)\n    - [Django Document and mapping](#django-document-and-mapping)\n        - [Custom declaration for fields](#custom-declaration-for-fields)\n        - [Excluding fields](#excluding-fields)\n        - [Custom mapping between serializer fields and ES fields](#custom-mapping-between-serializer-fields-and-es-fields)\n        - [Disabling the mapping](#disabling-the-mapping)\n        - [Relations](#relations)\n    - [Serializer](#serializer)\n    - [Objects and nested](#objects-and-nested)\n\n## Model and ES Document example\n\nTo declare a document, create a model and register it with a subclass of DjangoDocument. You do not need to declare any\nfields on the document - they are generated automatically.\n\n```python\nfrom django.db import models\nfrom django_es_drf import registry, DjangoDocument\n\n\nclass School(models.Model):\n    name = models.CharField(max_length=100)\n    address = models.TextField()\n\n\n@registry.register(School)\nclass SchoolDocument(DjangoDocument):\n    class Index:\n        name = "schools"\n```\n\nIf you want to split the file into models and documents, be free to do so but ensure that the file including documents\nis loaded at the startup - in app\'s ready function, imported at the bottom of the model etc.\n\nLater on, use `SchoolDocument` (which is a subclass of elasticsearch-dsl document):\n\n```python\ns = SchoolDocument(name=\'Blah\', address=\'somewhere\')\ns.save()\n\nfor s in SchoolDocument.search().filter(\'term\', name=\'Blah\'):\n    print(s)\n    s.name = s.name + \'-test\'\n    s.save()  # will save django and index to ES\n```\n\nYou can work with django objects if you prefer - they will get persisted into Elastic whenever they are saved:\n\n```python\ns = School.objects.create(name=\'blah\', address=\'somewhere\')\n# it was persisted, see:\ndoc = SchoolDocument.get(s.id)\nprint(doc.name)\n```\n\nYou can switch between django and document representations:\n\n```python\ndjango_school = School.objects.get(pk=1)\n\ndocument_school = SchoolDocument.from_django(django_school)\n\ndjango_again = document_school.to_django()\n\n# they are not the same instances though\nassert django_school is not django_again\n```\n\n## DRF example\n\nESViewSet is inherited from `ModelViewSet`, so feel free to use any of its features. You have to remember that\n`get_object` will give you DSL Document, not an instance of model.\n\nIn the viewset, specify a document, not a model/queryset.\n\n```python\nfrom django_es_drf import ESViewSet\n\n\nclass SchoolAPI(ESViewSet):\n    document = SchoolDocument\n```\n\nIf you need to pre-filter the initial queryset, override the `get_queryset`, call the parent and add your own initial\nfilter:\n\n```python\nfrom django_es_drf import ESViewSet\n\n\nclass SchoolAPI(ESViewSet):\n    document = SchoolDocument\n\n    def get_queryset(self):\n        qs = super().get_queryset()\n        return qs.filter(\'term\', cooperating=True)\n```\n\n### Returned fields from listing\n\nA `DynamicSourceFilter` is added automatically to filters and can be used to filter the properties returned in listings.\n\nYou can specify the props returned on the viewset:\n\n```python\nclass SchoolAPI(ESViewSet):\n    document = SchoolDocument\n    source = [\'id\', \'name\']\n```\n\nor alternatively, use `_include` or `_exclude` GET parameters containing a list of fields separated by commas\n(if `source` is set, they are active only on the fields from the source, otherwise from all fields)\n\nThe filter is only used for listing and listing-like operations (that is those where object id is not present in the\nURL). If you want to restrict the filter only on some calls, add a `is_dynamic_source_filtering_enabled(self, request)`\nmethod on the viewset class.\n\nExample:\n\n```\nGET /schools/?_exclude=address\n```\n\n### Search, facets\n\nDRF filter backends (that is `filter_backends` property on viewset) are used for both aggregations and search.\n\n#### Search via "q" GET parameters\n\nImplicitly `filter_backends` contain `QueryFilterBackend` that parses and interprets `q` and `parser` GET parameters:\n\n* the `parser` parameter contains name of the query interpreter that will be used. The interpreters are registered on \'\n  ESViewSet.query_interpreters\' dictionary which defaults\n  to ``{"simple": simple_query_interpreter, "luqum": luqum_query_interpreter}``. If the parameter is not passed,\n  ESViewSet.default_query_interpreter is used. If invalid value is passed, empty result list is returned.\n* `q` parameter is passed to the interpreter which applies it to a queryset\n\n##### Simple search\n\nThe simple interpreter performs `multi_match` operation on all fields.\n\n##### Luqum search\n\nThis search interpreter takes the content of the query and interprets it as a luqum query string. This library enables\nuse of `nested` mapping in query and supports boolean operators.\nSee [https://luqum.readthedocs.io/en/latest/about.html](https://luqum.readthedocs.io/en/latest/about.html)\nfor details on the language/capabilities.\n\n#### Facets and faceted filtering\n\nTo enable aggregations and filtering by bucket values, add `aggs` to the viewset class. Example:\n\n```python\nfrom django_es_drf import ESViewSet, BucketAgg\n\n\nclass SchoolAPI(ESViewSet):\n    document = SchoolDocument\n    aggs = [\n        BucketAgg("name"),\n        BucketAgg("address")\n    ]\n```\n\nYou can use the following aggregations:\n\n* `BucketAgg(name, field=None, type=\'terms\')` - a terms bucket filtering. If field is not set, it is the same as `name`.\n  Dot notation in the `field` is supported.\n* `NestedAgg(name, path, filter)` - generates a nested aggregation\n* `TranslatedBucketAgg(name, type="terms", field=None, map=None)` - translates the bucket keys through a map to get\n  human labels\n\nThe aggregations are chainable. Chaining `BucketAgg` does not mean nesting - the aggregations are on the same level.\n\nBigger Example:\n\n```python\naggs = [\n    TranslatedBucketAgg(\'available\', label=\'Available\', map={\n        True: \'yes\',\n        False: \'no\'\n    }),\n    NestedAgg("tagsWithYear")\n        .bucket("tagsWithYear.tag", label="Tag")\n        .bucket("tagsWithYear.year", label="Year"),\n]\n```\n\nAny options not interpreted by ES (in this case \'label\') are copied into the API output (and might be used, for example,\nin UI renderer).\n\nSample returned value (aggs = `aggs = [BucketAgg("name"), BucketAgg("address")]`):\n\n```json\n{\n  "aggs": [\n    {\n      "code": "name",\n      "__missing__": 0,\n      "__count__": 2,\n      "doc_count_error_upper_bound": 0,\n      "sum_other_doc_count": 0,\n      "buckets": [\n        {\n          "key": "test 1",\n          "doc_count": 1\n        },\n        {\n          "key": "test 2",\n          "doc_count": 1\n        }\n      ]\n    },\n    {\n      "code": "address",\n      "__missing__": 0,\n      "__count__": 2,\n      "doc_count_error_upper_bound": 0,\n      "sum_other_doc_count": 0,\n      "buckets": [\n        {\n          "key": "blah",\n          "doc_count": 2\n        }\n      ]\n    }\n  ],\n  ...\n}\n```\n\n##### Filtering by aggregation buckets\n\nTo filter by aggregation buckets, add `?f=<bucket_code>:<value>:<bucket_code>:<value>...` to the url. URLEncode \'%\'\nand \':\' if there is \':\' in the unlikely case these chars are in the value. If you need a different syntax, define\na `parse_facet_query(self, request)` method on you viewset that returns either None or dictionary with facet code as a\nkey and a list of values to match as the value.\n\nExample:\n\n```GET /schools/?f=name:test 2```\n\nResponse:\n\n```json5\n            {\n  "count": 1,\n  // ...\n  "hits": [\n    {\n      "id": 2,\n      "name": "test 2",\n      "address": "blah"\n    }\n  ],\n  "aggs": [\n    // ...\n  ]\n}\n```\n\n##### Filtering without aggregations\n\nA bucket aggregation can be used just to filter the queryset, not to generate any aggregations. To do so, specify\n`display=False` option to the aggregation. The `?f` will still be available, but no aggregations will be generated.\n\n```python\naggs = [\n    BucketAgg(\'available\', display=False),\n    NestedAgg("tagsWithYear")\n        .bucket("tagsWithYear.tag", label="Tag")\n        .bucket("tagsWithYear.year", label="Year"),\n]\n```\n\n## Django Document and mapping\n\n### Custom declaration for fields\n\nAny fields that are already present on the document will be keep as they are. In the example above, to set that\n`name` is a `text` and not `keyword` (the default), just declare:\n\n```python\nimport elasticsearch_dsl as e\n\n\n@registry.register(School)\nclass SchoolDocument(DjangoDocument):\n    name = e.Text()\n\n    class Index:\n        name = "schools"\n```\n\n### Excluding fields\n\nTo exclude a field from the automatic generation, add it to `excluded`. If "nested/object" mapping will be generated,\nyou can use a dot-separated path.\n\n```python\n@registry.register(School, excluded=[\'address\'])\nclass SchoolDocument(DjangoDocument):\n    class Index:\n        name = "schools"\n```\n\n### Custom mapping between serializer fields and ES fields\n\nAdd your own mapping - the key is the DRF field type, value is a function that takes the field and context and returns\nES field:\nThe context is an instance of `RegistrationContext`.\n\n```python\nimport elasticsearch_dsl as e\nimport rest_framework.serializers as s\n\n\n@registry.register(School,\n                   mapping={\n                       s.CharField: lambda fld, ctx, **kwargs: e.Keyword()\n                   })\nclass SchoolDocument(DjangoDocument):\n    class Index:\n        name = "schools"\n```\n\n### Disabling the mapping\n\nAdd `generate=False` to decorator\'s parameters:\n\n```python\nimport elasticsearch_dsl as e\n\n\n@registry.register(School,\n                   generate=False)\nclass SchoolDocument(DjangoDocument):\n    # you need to provide your own mapping here\n\n    class Index:\n        name = "schools"\n```\n\n### Relations\n\nThe framework has a rudimentary support for read-only relations. If there is an M2M or ForeignKey on model and\nserializer uses target\'s serializer to embed the representation of the relation target, the document will contain an\nobject mapping for the target\'s data. It can be changed to Nested via mapping parameter in Document\'s registration\ndecorator.\n\nThe framework does not provide any support for propagating changes on related models. If you need this, write your own\nchange listeners or use a more complete library, such as django-elasticsearch-dsl-drf.\n\nA simple example with a foreign key:\n\n```python\nclass City(models.Model):\n    name = models.CharField(max_length=100)\n\n\nclass School(models.Model):\n    name = models.CharField(max_length=100)\n    city = models.ForeignKey(City, related_name="+", on_delete=models.CASCADE)\n\n\n@registry.register(School, serializer_meta={"depth": 1})\nclass SchoolDocument(DjangoDocument):\n    class Index:\n        name = "schools"\n```\n\n*Note:* the `depth` in `serializer_meta` argument - it instructs the generated model serializer to create a serializer\nfor city as well. This is roughly equivalent to the following code:\n\n```python\nfrom rest_framework import serializers\n\n\nclass CitySerializer(serializers.ModelSerializer):\n    class Meta:\n        model = City\n        exclude = ()\n\n\nclass SchoolSerializer(serializers.ModelSerializer):\n    city = CitySerializer()\n\n    class Meta:\n        model = School\n        exclude = ()\n\n\n@registry.register(School, serializer=SchoolSerializer)\nclass SchoolDocument(DjangoDocument):\n    class Index:\n        name = "schools"\n```\n\nIf you want to change the type to nested:\n\n```python\nimport elasticsearch_dsl as e\n\n\n@registry.register(School, serializer_meta={"depth": 1}, mapping={\n    \'city\': e.Nested\n})\nclass SchoolDocument(DjangoDocument):\n    class Index:\n        name = "schools"\n\n```\n\n*Note:* There is no support for creating nested objects. If you need this support, you\'ll have to write your\nown `create`/`update` method on the serializer.\n\n## Serializer\n\nOptionally you might want to provide your own serializer, to transform/generate additional fields into the document.\n\nThe serializer is just a plain DRF serializer that converts django fields to document\'s fields, there is nothing fancy\nabout it.\n\n*Note:* If you use `SerializerMethodField`, be sure to add the correct field into the mapping:\n\n```python\nfrom rest_framework import serializers\n\n\nclass SchoolSerializer(serializers.ModelSerializer):\n    name_with_address = serializers.SerializerMethodField()\n\n    def get_name_with_address(self, instance):\n        return f"{instance.name}, {instance.address}"\n\n    class Meta:\n        model = School\n        exclude = ()\n\n\n@registry.register(School, serializer=SchoolSerializer, mapping={\n    \'name_with_address\': e.Text\n})\nclass SchoolDocument(DjangoDocument):\n    class Index:\n        name = "schools"\n```\n',
    'author': 'Mirek Simek',
    'author_email': 'miroslav.simek@vscht.cz',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



from .document_registry import registry, DjangoDocument
from .drf.viewsets import ESViewSet
from .drf.aggs import AggBase, BucketAgg, TranslatedBucketAgg, NestedAgg
from .drf.serializers import ESDocumentSerializer
from .drf.pagination import ESPagination, ESPage, ESPaginator
from .drf.backends.source import DynamicSourceBackend
from .drf.backends.filters import (
    ESAggsFilterBackend,
    QueryFilterBackend,
    BaseESFilterBackend,
)
from .drf.renderers import ESRenderer, ESEncoderClass
